# -*- coding: utf-8 -*-
import re

import six

from decorated.util import dutil

if six.PY2:
    _DEFAULT_STRING_TYPE = 'unicode'
else:
    _DEFAULT_STRING_TYPE = 'str'
    
class Template(object):
    def __init__(self, string):
        self._string = string
        source = _generate_source(string)
        self._code = compile(source, '<string>', 'exec')
        
    def __call__(_template_self, **variables): # variables may contain a "self" key
        variables = dutil.generate_safe_context(variables)
        exec(_template_self._code, variables)
        return variables['result']
    
    def __str__(self):
        return self._string

def _generate_source(string):
    parts = re.split('(\{.+?\})', string)
    for i, part in enumerate(parts):
        if part == '':
            parts[i] = None
        elif part.startswith('{') and part.endswith('}'):
            part = part[1:-1]
            part = '''
try:
    parts.append(%s(%s))
except Exception:
    parts.append(u'{error:%s}')
''' % (_DEFAULT_STRING_TYPE, part, part)
            parts[i] = part.strip()
        else:
            parts[i] = "parts.append(u'%s')" % part.replace('\n', '\\n').replace("'", "\\'")
    parts = '\n'.join([p for p in parts if p is not None])
    source = '''
parts = []
%s
result = ''.join(parts)
''' % parts
    return source.strip()



# -*- coding: utf-8 -*-
from decorated import WrapperFunction
from decorated.decorators.validations.errors import ValidationError
from decorated.util import dutil


class ValidationEngine(object):
    def __init__(self, default_error_class=ValidationError):
        self._default_error_class = default_error_class

    def rules(self, validators, error_class=None):
        error_class = error_class or self._default_error_class
        return Rules(validators, error_class)

class Rules(WrapperFunction):
    def _before(self, *args, **kw):
        arg_dict = self._resolve_args(*args, **kw)
        for validator in self._validators:
            validator.validate(arg_dict, default_error_class=self._error_class)

    def _init(self, validators, error_class=None):
        super(Rules, self)._init()
        self._validators = dutil.listify(validators)
        self._error_class = error_class
        


# -*- coding: utf-8 -*-
from cProfile import Profile as RawProfile
from decorated.base.function import Function
from decorated.util import reporters
from decorated.util.gcutil import DisableGc
from pstats import Stats as OriginalStats
from six import StringIO
import doctest

class Profile(Function):
    def _call(self, *args, **kw):
        profile = RawProfile()
        def _run():
            with DisableGc():
                for _ in range(self._iterations):
                    _run.result = super(Profile, self)._call(*args, **kw)
        profile.runctx('_run()', {}, {'_run': _run})
        profile.create_stats()
        stats = Stats(profile)
        stats.sort_stats('cumulative')
        stats.fcn_list = stats.fcn_list[:self._max_lines] # pylint: disable=attribute-defined-outside-init
        self._reporter(stats)
        return _run.result
    
    def _init(self, iterations=1, reporter=reporters.PRINT_REPORTER, max_lines=50): # pylint: disable=arguments-differ
        super(Profile, self)._init()
        self._iterations = iterations
        self._reporter = reporter
        self._max_lines = max_lines
        
class Stats(OriginalStats):
    def __str__(self):
        '''
        >>> result = str(Stats(RawProfile().run('')))
        >>> 'function calls' in result
        True
        '''
        self.stream = StringIO()
        self.print_stats()
        return self.stream.getvalue()



# -*- coding: utf-8 -*-
import logging
import sys
import time

import six

from decorated.base.function import Function

log = logging.getLogger(__name__)

class Retries(Function):
    def _init(self, times, delay=0, error_types=(Exception,)): # pylint: disable=arguments-differ
        if times < 0:
            raise Exception('Times cannot be negative.')
        self._times = times
        self._delay = delay
        self._error_types = error_types
        
    def _call(self, *args, **kw):
        for i in six.moves.xrange(self._times + 1):
            try:
                return super(Retries, self)._call(*args, **kw)
            except Exception as e:
                if i < self._times and isinstance(e, self._error_types):
                    log.warn('Execution failed. Will retry in %f seconds.', self._delay, exc_info=True)
                    time.sleep(self._delay)
                else:
                    raise

class RetriesForever(Retries):
    def _init(self, delay=0, error_types=(Exception,)):
        super(RetriesForever, self)._init(sys.maxint - 1, delay=delay, error_types=error_types)



# -*- coding: utf-8 -*-
from decimal import Decimal

import six

from decorated.decorators.validations.validators.misc import TypeValidator

_NUMBER_TYPES = list(six.integer_types) + [float, Decimal]
try:
    from cdecimal import Decimal as CDecimal
    _NUMBER_TYPES.append(CDecimal)
except Exception:
    pass
_NUMBER_TYPES = tuple(_NUMBER_TYPES)

class NumberValidator(TypeValidator):
    def __init__(self, param, error_class=None):
        super(NumberValidator, self).__init__(param, _NUMBER_TYPES, error_class=error_class)

class BetweenValidator(NumberValidator):
    def __init__(self, param, lower, upper, error_class=None):
        super(BetweenValidator, self).__init__(param, error_class=error_class)
        self._lower = lower
        self._upper = upper

    def _validate(self, value):
        '''
        >>> BetweenValidator('score', 1, 10)._validate(5)
        >>> BetweenValidator('score', 1, 10)._validate(0)
        'should be between 1 and 10'
        >>> BetweenValidator('score', 1, 10)._validate(11)
        'should be between 1 and 10'
        '''
        if value < self._lower or value > self._upper:
            return 'should be between %s and %s' % (self._lower, self._upper)

class GeValidator(NumberValidator):
    def __init__(self, param, threshold, error_class=None):
        super(GeValidator, self).__init__(param, error_class=error_class)
        self._threshold = threshold
    
    def _validate(self, value):
        '''
        >>> GeValidator('value', 0)._validate(1)
        >>> GeValidator('value', 0)._validate(0)
        >>> GeValidator('value', 0)._validate(-1)
        'should be >=0'
        '''
        if value < self._threshold:
            return 'should be >=%s' % self._threshold

class GtValidator(NumberValidator):
    def __init__(self, param, threshold, error_class=None):
        super(GtValidator, self).__init__(param, error_class=error_class)
        self._threshold = threshold
        
    def _validate(self, value):
        '''
        >>> GtValidator('value', 0)._validate(1)
        >>> GtValidator('value', 0)._validate(0)
        'should be >0'
        >>> GtValidator('value', 0)._validate(-1)
        'should be >0'
        '''
        if value <= self._threshold:
            return 'should be >%s' % self._threshold

class LeValidator(NumberValidator):
    def __init__(self, param, threshold, error_class=None):
        super(LeValidator, self).__init__(param, error_class=error_class)
        self._threshold = threshold
    
    def _validate(self, value):
        '''
        >>> LeValidator('value', 0)._validate(-1)
        >>> LeValidator('value', 0)._validate(0)
        >>> LeValidator('value', 0)._validate(1)
        'should be <=0'
        '''
        if value > self._threshold:
            return 'should be <=%s' % self._threshold

class LtValidator(NumberValidator):
    def __init__(self, param, threshold, error_class=None):
        super(LtValidator, self).__init__(param, error_class=error_class)
        self._threshold = threshold
    
    def _validate(self, value):
        '''
        >>> LtValidator('value', 0)._validate(-1)
        >>> LtValidator('value', 0)._validate(0)
        'should be <0'
        >>> LtValidator('value', 0)._validate(1)
        'should be <0'
        '''
        if value >= self._threshold:
            return 'should be <%s' % self._threshold

class PositiveValidator(NumberValidator):
    def _validate(self, value):
        '''
        >>> PositiveValidator('score')._validate(5)
        >>> PositiveValidator('score')._validate(-1)
        'should be positive'
        '''
        if value < 0:
            return 'should be positive'

class NonNegativeValidator(NumberValidator):
    def _validate(self, value):
        '''
        >>> NonNegativeValidator('score')._validate(5)
        >>> NonNegativeValidator('score')._validate(0)
        'should be non negative'
        '''
        if value <= 0:
            return 'should be non negative'
        


# -*- coding: utf-8 -*-
# pylint: disable=invalid-name
from decorated.base.false import NOTSET
from decorated.base.context import Context, ctx
from decorated.base.function import ContextFunction, Function, WrapperFunction, partial
from decorated.decorators.conditional import Conditional
from decorated.decorators.events import Event, event
from decorated.decorators.files import TempDir, TempFile, WritingFile
from decorated.decorators.once import Once, OnceSession
from decorated.decorators.profile import Profile
from decorated.decorators.remove_extra_args import RemoveExtraArgs
from decorated.decorators.retries import Retries, RetriesForever
from decorated.decorators.timeit import TimeIt
from decorated.decorators.timeout import Timeout

conditional = Conditional
once = Once
OnceSession = OnceSession
profile = Profile
remove_extra_args = RemoveExtraArgs
retries = Retries
retries_forever = RetriesForever
tempfile = TempFile
tempdir = TempDir
timeit = TimeIt
timeout = Timeout
writing_file = WritingFile



#!/usr/bin/python
from decorated.decorators.validations.engine import ValidationEngine
from decorated.decorators.validations.validators.misc import ChoicesValidator, NotEmptyValidator, NotNoneValidator, \
    TypeValidator
from decorated.decorators.validations.validators.number import BetweenValidator, NonNegativeValidator, NumberValidator, \
    PositiveValidator, GeValidator, GtValidator, LeValidator, LtValidator
from decorated.decorators.validations.validators.string import MaxLengthValidator, RegexValidator

engine = ValidationEngine()

between = BetweenValidator
choices = ChoicesValidator
ge = GeValidator
gt = GtValidator
le = LeValidator
lt = LtValidator
max_length = MaxLengthValidator
not_empty = NotEmptyValidator
not_none = NotNoneValidator
non_negative = NonNegativeValidator
number = NumberValidator
positive = PositiveValidator
regex = RegexValidator
type = TypeValidator



# -*- coding: utf-8 -*-
from collections import Iterable

import six

_SAFE_BUILTINS = {
    '__builtins__': None,
    'Exception': Exception,
    'False': False,
    'None': None,
    'True': True,
    'float': float,
    'int': int,
    'isinstance': isinstance,
    'issubclass': issubclass,
    'list': list,
    'len': len,
    'max': max,
    'min': min,
    'round': round,
    'sorted': sorted,
    'str': str,
    'sum': sum,
    'tuple': tuple,
    'type': type,
}
if six.PY2:
    _SAFE_BUILTINS['unicode'] = unicode

def generate_safe_context(variables):
    ctx = dict(variables)
    ctx.update(_SAFE_BUILTINS)
    return ctx

def listify(value_or_values):
    '''
    >>> listify(['aaa', 'bbb'])
    ['aaa', 'bbb']
    >>> listify(('aaa', 'bbb'))
    ['aaa', 'bbb']
    >>> listify('aaa')
    ['aaa']
    '''
    if isinstance(value_or_values, list):
        return value_or_values
    elif isinstance(value_or_values, six.string_types):
        return [value_or_values]
    elif isinstance(value_or_values, Iterable):
        return list(value_or_values)
    else:
        return [value_or_values]



# -*- coding: utf-8 -*-

class Validator(object):
    def __init__(self, param, error_class=None):
        self._param = param
        self._error_class = error_class
        self._validator_classes = []
        self._validator_classes = _list_validator_classes(self.__class__)

    def validate(self, arg_dict, default_error_class=None):
        error_class = self._error_class or default_error_class
        value = self._eval_value(arg_dict, error_class)
        for cls in self._validator_classes:
            violation = cls._validate(self, value)
            if violation is not None:
                e = error_class('Arg "%s" %s, got "%s" (type=%s).' % (self._param, violation, value, type(value).__name__))
                e.param = self._param
                raise e

    def _eval_value(self, arg_dict, error_class):
        if callable(self._param):
            try:
                value = self._param(**arg_dict)
            except Exception:
                raise error_class('Arg "%s" is missing or malformed.' % self._param)
        else:
            try:
                value = arg_dict[self._param]
            except KeyError:
                raise error_class('Arg "%s" is missing.' % self._param)
        return value

    def _validate(self, value):
        raise NotImplementedError()
    
def _list_validator_classes(cls):
    '''
    >>> class ParentValidator(Validator): pass
    >>> class ChildValidator(ParentValidator): pass
    >>> class GrandChildValidator(ChildValidator): pass
    >>> classes = _list_validator_classes(GrandChildValidator)
    >>> [c.__name__ for c in classes]
    ['ParentValidator', 'ChildValidator', 'GrandChildValidator']
    '''
    classes = []
    while issubclass(cls, Validator) and cls is not Validator:
        classes.append(cls)
        cls = cls.__bases__[0]
    return tuple(reversed(classes))



# -*- coding: utf-8 -*-
import six

from decorated.base.context import ContextError, ContextMeta
from decorated.base.expression import Expression
from decorated.base.false import NOTSET
from decorated.base.function import Function

class Once(Function):
    def _call(self, *args, **kw):
        arg_dict = self._resolve_args(*args, **kw)
        key = self._key(**arg_dict)
        key = (self._func, key)

        try:
            values = OnceSession.current().values
        except ContextError:
            values = _DEFAULT_SESSION.values
        except AttributeError:
            values = _DEFAULT_SESSION.values

        value = values.get(key, NOTSET)
        if value == NOTSET:
            value = super(Once, self)._call(*args, **kw)
            values[key] = value
        return value
    
    def _init(self, key='None'): # pylint: disable=arguments-differ
        super(Once, self)._init()
        self._key = Expression(key)
        
class OnceSession(six.with_metaclass(ContextMeta, object)):
    def __init__(self):
        super(OnceSession, self).__init__()
        self.values = {}
        
_DEFAULT_SESSION = OnceSession()



# -*- coding: utf-8 -*-
from decorated.util import dutil


class Expression(object):
    def __init__(self, string):
        self._string = string
        try:
            self._expression = compile(string, '<string>', 'eval')
        except SyntaxError:
            raise ExpressionError('Bad expression "%s".' % string)
        
    def __call__(_expression_self, **variables): # variables may contain a "self" key
        variables = dutil.generate_safe_context(variables)
        try:
            return eval(_expression_self._expression, variables)
        except Exception as e:
            raise ExpressionError('Failed to evaluate expression "%s". Error was: %s' % (_expression_self._string, e))
        
    def __str__(self):
        return self._string

class ExpressionError(Exception):
    pass



# -*- coding: utf-8 -*-
from collections import Iterable

from decorated.decorators.validations.validators.base import Validator


class ChoicesValidator(Validator):
    def __init__(self, param, choices, error_class=None):
        super(ChoicesValidator, self).__init__(param, error_class=error_class)
        self._choices = choices

    def _validate(self, value):
        '''
        >>> ChoicesValidator('id', [1, 2, 3])._validate(2)
        >>> ChoicesValidator('id', [1, 2, 3])._validate(0)
        'should be one of [1, 2, 3]'
        '''
        if value not in self._choices:
            return 'should be one of %s' % self._choices

class NotNoneValidator(Validator):
    def _validate(self, value):
        '''
        >>> NotNoneValidator('id')._validate(111)
        >>> NotNoneValidator('id')._validate(None)
        'should not be none'
        '''
        if value is None:
            return 'should not be none'

class NotEmptyValidator(Validator):
    def _validate(self, value):
        '''
        >>> NotEmptyValidator('values')._validate(111)
        >>> NotEmptyValidator('values')._validate(True)
        >>> NotEmptyValidator('values')._validate(False)
        >>> NotEmptyValidator('values')._validate(None)
        'should not be empty'
        >>> NotEmptyValidator('values')._validate([])
        'should not be empty'
        >>> NotEmptyValidator('values')._validate('')
        'should not be empty'
        '''
        if not value and not isinstance(value, bool):
            return 'should not be empty'

class TypeValidator(Validator):
    def __init__(self, param, types, error_class=None):
        super(TypeValidator, self).__init__(param, error_class=error_class)
        self._types = types
        if isinstance(types, Iterable):
            self._types_string = '/'.join([t.__name__ for t in types])
        else:
            self._types_string = types.__name__

    def _validate(self, value):
        '''
        >>> TypeValidator('value', int)._validate(111)
        >>> TypeValidator('value', int)._validate('111')
        'should be int'
        >>> TypeValidator('value', (int, float))._validate('111')
        'should be int/float'
        '''
        if not isinstance(value, self._types):
            return 'should be %s' % self._types_string
        


# -*- coding: utf-8 -*-
import re

import six

from decorated.decorators.validations.validators.misc import TypeValidator


class StringValidator(TypeValidator):
    def __init__(self, param, error_class=None):
        super(StringValidator, self).__init__(param, six.string_types, error_class=error_class)
        
class MaxLengthValidator(StringValidator):
    def __init__(self, param, max_length, error_class=None):
        super(MaxLengthValidator, self).__init__(param, error_class=error_class)
        self._max_length = max_length

    def _validate(self, value):
        '''
        >>> MaxLengthValidator('name', 8)._validate('12345')
        >>> MaxLengthValidator('name', 8)._validate('123456789')
        'should be less than 8 chars'
        '''
        if len(value) > self._max_length:
            return 'should be less than %d chars' % self._max_length

class RegexValidator(StringValidator):
    def __init__(self, param, regex, error_class=None):
        super(RegexValidator, self).__init__(param, error_class=error_class)
        self._regex = regex
        self._compiled_regex = re.compile(regex)

    def _validate(self, value):
        '''
        >>> RegexValidator('name', '[a-z]+')._validate('aaa')
        >>> RegexValidator('name', '[a-z]+')._validate('111')
        'should match regex "[a-z]+"'
        '''
        if not self._compiled_regex.match(value):
            return 'should match regex "%s"' % self._regex
        


# -*- coding: utf-8 -*-
import time

from decorated.base.function import Function
from decorated.base.template import Template
from decorated.util import modutil


class _CacheDecorator(Function):
    @property
    def invalidate(self):
        return self._cache.uncache(self._key, **self._vars)

    def _call(self, *args, **kw):
        arg_dict = dict(self._vars)
        arg_dict.update(self._resolve_args(*args, **kw))
        key = self._key(**arg_dict)
        return self._process(key, *args, **kw)

    def _init(self, cache, key, vars=None, options=None): # pylint: disable=arguments-differ
        super(_CacheDecorator, self)._init()
        self._cache = cache
        self._key = key if callable(key) else Template(key)
        self._vars = vars or {}
        self._options = options

    def _process(self, key, *args, **kw):
        raise NotImplementedError()

class Cache(object):
    def cache(self, key, vars=None, **options):
        class _Decorator(_CacheDecorator):
            def _process(self, key, *args, **kw):
                result = self._cache.get(key, **options)
                if result is None:
                    result = self._func(*args, **kw)
                    self._cache.set(key, result, **options)
                return result
        return _Decorator(self, key, vars=vars, options=options)

    def clear(self):
        raise NotImplementedError()
    
    def delete(self, key, **options):
        raise NotImplementedError()
    
    def get(self, key, **options):
        raise NotImplementedError()
    
    def set(self, key, value, **options):
        raise NotImplementedError()

    def uncache(self, key, vars=None, **options):
        class _Decorator(_CacheDecorator):
            def _process(self, key, *args, **kw):
                self._cache.delete(key, **options)
                return self._func(*args, **kw)
        return _Decorator(self, key, vars=vars, options=options)

class SimpleCache(Cache):
    def __init__(self):
        super(SimpleCache, self).__init__()
        self.clear()

    def clear(self):
        self._data = {}
        
    def delete(self, key, **options):
        del self._data[key]
        
    def get(self, key, **options):
        value, expires = self._data.get(key, (None, None))
        if expires is None or expires < time.time():
            return None
        else:
            return value
        
    def set(self, key, value, **options):
        expires = time.time() + options.get('ttl', 3600)
        self._data[key] = (value, expires)
        
if modutil.module_exists('pylru'):
    from pylru import lrucache
    
    class LruCache(Cache):
        def __init__(self, size=1000):
            self._size = size
            self.clear()

        def clear(self):
            self._cache = lrucache(self._size)
            
        def delete(self, key, **options):
            del self._cache[key]
        
        def get(self, key, **options):
            try:
                return self._cache[key]
            except KeyError:
                return None
        
        def set(self, key, value, **options):
            self._cache[key] = value
            


# -*- coding: utf-8 -*-
import fcntl
import os
from threading import RLock

from decorated import ContextFunction, WrapperFunction


class Lock(ContextFunction):
    def lock(self):
        raise NotImplementedError()

    def unlock(self):
        raise NotImplementedError()

    def _after(self, ret, *args, **kw):
        self.unlock()

    def _before(self, *args, **kw):
        self.lock()

    def _decorate(self, func):
        return _LockProxy(self)(func)

    def _error(self, error, *args, **kw):
        self.unlock()

class FileLock(Lock):
    def _init(self, path): # pylint: disable=arguments-differ
        super(FileLock, self)._init()
        self._path = path
        self._fd = None

    def lock(self):
        _create_file_if_not_exist(self._path)
        self._fd = os.open(self._path, os.O_RDWR)
        fcntl.flock(self._fd, fcntl.LOCK_EX)

    def unlock(self):
        fcntl.flock(self._fd, fcntl.LOCK_UN)

class MemoryLock(Lock):
    def _init(self, *args, **kw):
        super(MemoryLock, self)._init(*args, **kw)
        self._lock = RLock()

    def lock(self):
        self._lock.acquire()

    def unlock(self):
        self._lock.release()

class _LockProxy(WrapperFunction):
    def _after(self, ret, *args, **kw):
        self._target.unlock()

    def _before(self, *args, **kw):
        self._target.lock()

    def _error(self, error, *args, **kw):
        self._target.unlock()

    def _init(self, target): # pylint: disable=arguments-differ
        super(_LockProxy, self)._init()
        self._target = target

    def _is_init_args(self, *args, **kw):
        return True

def _create_file_if_not_exist(path):
    try:
        f = open(path, 'w')
        f.close()
    except OSError:
        pass



# -*- coding: utf-8 -*-

class FalseObject(object):
    '''
    >>> bool(FalseObject())
    False
    '''

    def __bool__(self):
        return False

    def __nonzero__(self):
        return False

NOTSET = FalseObject()
